@@ -15,6 +15,7 @@ from numpy.random import RandomState
 
 from rlai.actions import Action, DiscretizedAction, ContinuousMultiDimensionalAction
 from rlai.agents.mdp import MdpAgent
+from rlai.environments import Environment
 from rlai.environments.mdp import ContinuousMdpEnvironment
 from rlai.meta import rl_text
 from rlai.models.feature_extraction import (
@@ -40,8 +41,8 @@ class GymState(MdpState):
 
     def __init__(
             self,
-            environment,
-            observation,
+            environment: 'Gym',
+            observation: np.ndarray,
             agent: MdpAgent,
             terminal: bool,
     ):
@@ -54,8 +55,6 @@ class GymState(MdpState):
         :param terminal: Whether the state is terminal.
         """
 
-        environment: Gym
-
         super().__init__(
             i=agent.pi.get_state_i(observation),
             AA=environment.actions,
@@ -72,10 +71,13 @@ class Gym(ContinuousMdpEnvironment):
     """
 
     LLC_V2 = 'LunarLanderContinuous-v2'
+    LLC_V2_FUEL_CONSUMPTION_FULL_THROTTLE_MAIN = 1 / 300.0
+    LLC_V2_FUEL_CONSUMPTION_FULL_THROTTLE_SIDE = 1 / 600.0
+
     MCC_V0 = 'MountainCarContinuous-v0'
-    MMC_V0_TROUGH_X_POS = -0.5
-    MMC_V0_GOAL_X_POS = 0.45
-    MMC_V0_FUEL_CONSUMPTION_FULL_THROTTLE = 1.0 / 300.0
+    MCC_V0_TROUGH_X_POS = -0.5
+    MCC_V0_GOAL_X_POS = 0.45
+    MCC_V0_FUEL_CONSUMPTION_FULL_THROTTLE = 1.0 / 300.0
 
     @classmethod
     def get_argument_parser(
@@ -136,6 +138,12 @@ class Gym(ContinuousMdpEnvironment):
             help='Pass this flag to plot environment values (e.g., state).'
         )
 
+        parser.add_argument(
+            '--progressive-reward',
+            action='store_true',
+            help=f'Pass this flag to use progressive rewards (only valid for {Gym.MCC_V0}).'
+        )
+
         return parser
 
     @classmethod
@@ -143,7 +151,7 @@ class Gym(ContinuousMdpEnvironment):
             cls,
             args: List[str],
             random_state: RandomState
-    ) -> Tuple[ContinuousMdpEnvironment, List[str]]:
+    ) -> Tuple[Environment, List[str]]:
         """
         Initialize an environment from arguments.
 
@@ -181,42 +189,86 @@ class Gym(ContinuousMdpEnvironment):
         # map discretized actions back to continuous space
         if isinstance(a, DiscretizedAction):
             gym_action = a.continuous_value
-        # use continuous action values directly (need to be arrays)
+        # use continuous action values (which are vectors) directly
         elif isinstance(a, ContinuousMultiDimensionalAction):
             gym_action = a.value
         # use discretized action indices
         else:
             gym_action = a.i
 
-        # continuous mountain car:  cap energy expenditure at fuel level
-        if self.gym_id == Gym.MCC_V0:
-            fuel_level = state.observation[2]
-            fuel_max_throttle = fuel_level / Gym.MMC_V0_FUEL_CONSUMPTION_FULL_THROTTLE
-            desired_throttle = gym_action[0]
-            effective_throttle = min(fuel_max_throttle, abs(desired_throttle))
-            gym_action[0] = np.sign(desired_throttle) * effective_throttle
+        # fuel-based modification for continuous environments. cap energy expenditure at remaining fuel levels.
+        fuel_used = None
+        if self.gym_id == Gym.LLC_V2:
+            main_throttle, side_throttle = gym_action[:]
+            required_main_fuel = Gym.LLC_V2_FUEL_CONSUMPTION_FULL_THROTTLE_MAIN * (0.5 + 0.5 * main_throttle if main_throttle >= 0.0 else 0.0)
+            required_side_fuel = Gym.LLC_V2_FUEL_CONSUMPTION_FULL_THROTTLE_SIDE * abs(side_throttle) if abs(side_throttle) >= 0.5 else 0.0
+            required_total_fuel = required_main_fuel + required_side_fuel
+            fuel_level = state.observation[-1]
+            if required_total_fuel > fuel_level:  # pragma no cover
+                gym_action[:] *= fuel_level / required_total_fuel
+                fuel_used = fuel_level
+            else:
+                fuel_used = required_total_fuel
+
+        elif self.gym_id == Gym.MCC_V0:
+            throttle = gym_action[0]
+            required_fuel = Gym.MCC_V0_FUEL_CONSUMPTION_FULL_THROTTLE * abs(throttle)
+            fuel_level = state.observation[-1]
+            if required_fuel > fuel_level:  # pragma no cover
+                gym_action[:] *= fuel_level / required_fuel
+                fuel_used = fuel_level
+            else:
+                fuel_used = required_fuel
 
         observation, reward, done, _ = self.gym_native.step(action=gym_action)
 
-        # override reward per environment. continuous lunar lander:  the ideal landing is zeros across the board.
-        if self.gym_id == Gym.LLC_V2:
-            reward = -np.abs(observation[0:6]).sum() if done else 0.0
-
-        # continuous mountain car
-        elif self.gym_id == Gym.MCC_V0:
-
-            # append fuel level to state
-            fuel_consumed = abs(gym_action[0]) * Gym.MMC_V0_FUEL_CONSUMPTION_FULL_THROTTLE
-            fuel_level = state.observation[2]
-            fuel_remaining = max(0.0, fuel_level - fuel_consumed)
+        # update fuel remaining if needed
+        fuel_remaining = None
+        if fuel_used is not None:
+            fuel_remaining = max(0.0, state.observation[-1] - fuel_used)
             observation = np.append(observation, fuel_remaining)
 
-            fraction_to_goal = (observation[0] - Gym.MMC_V0_TROUGH_X_POS) / (Gym.MMC_V0_GOAL_X_POS - Gym.MMC_V0_TROUGH_X_POS)
+        if self.gym_id == Gym.LLC_V2:
 
-            # reward at apex of the climb
-            reward = fraction_to_goal + fuel_remaining if fraction_to_goal >= 1.0 else 0.0
-            if self.previous_observation is not None and self.previous_observation[1] > 0.0 and observation[1] <= 0.0 and observation[0] >= Gym.MMC_V0_TROUGH_X_POS:
-                reward = fuel_remaining * fraction_to_goal
+            reward = 0.0
+
+            if done:
+
+                # the ideal state is zeros across position/movement
+                state_reward = -np.abs(observation[0:6]).sum()
+
+                # reward for remaining fuel, but only if the state is good. rewarding for remaining fuel unconditionally
+                # can cause the agent to veer out of bounds immediately and thus sacrifice state reward for fuel reward.
+                # the terminating state is considered good if the lander is within the goal posts (which are at
+                # x = +/-0.2) and the other orientation variables (y position, x and y velocity, angle and angular
+                # velocity) are near zero. permit a small amount of lenience in the latter, since it's common for a
+                # couple of the variables to be slightly positive even when the lander is sitting stationary on a flat
+                # surface.
+                fuel_reward = 0.0
+                if abs(observation[0]) <= 0.2 and np.abs(observation[1:6]).sum() < 0.01:  # pragma no cover
+                    fuel_reward = state.observation[-1]
+
+                reward = state_reward + fuel_reward
+
+        elif self.gym_id == Gym.MCC_V0:
+
+            reward = 0.0
+
+            # calculate fraction to goal state
+            curr_distance = observation[0] - Gym.MCC_V0_TROUGH_X_POS
+            goal_distance = self.mcc_curr_goal_x_pos - Gym.MCC_V0_TROUGH_X_POS
+            fraction_to_goal = curr_distance / goal_distance
+            if fraction_to_goal >= 1.0:
+
+                # increment goal up to the final goal
+                self.mcc_curr_goal_x_pos = min(Gym.MCC_V0_GOAL_X_POS, self.mcc_curr_goal_x_pos + 0.05)
+
+                # mark state and stats recorder as done. must manually mark stats recorder to allow premature reset.
+                done = True
+                if hasattr(self.gym_native, 'stats_recorder'):
+                    self.gym_native.stats_recorder.done = done
+
+                reward = curr_distance + fuel_remaining
 
         # call render if rendering manually
         if self.check_render_current_episode(True):
@@ -260,8 +312,8 @@ class Gym(ContinuousMdpEnvironment):
 
         observation = self.gym_native.reset()
 
-        # append fuel level to state of continuous mountain car
-        if self.gym_id == Gym.MCC_V0:
+        # append fuel level to state of certain continuous environments
+        if self.gym_id in [Gym.MCC_V0, Gym.LLC_V2]:
             observation = np.append(observation, 1.0)
 
         # call render if rendering manually
@@ -284,7 +336,7 @@ class Gym(ContinuousMdpEnvironment):
         """
         Check whether the current episode is to be rendered.
 
-        :param render_manually: Wether the rendering will be done manually with calls to the render function or automatically
+        :param render_manually: Whether the rendering will be done manually with calls to the render function or automatically
         as a result of saving videos via the monitor. Pass None to check whether the episode should be rendered,
         regardless of how the rendering will be done.
         :return: True if rendered and False otherwise.
@@ -377,7 +429,8 @@ class Gym(ContinuousMdpEnvironment):
                 'ang',
                 'angV',
                 'leg1Con',
-                'leg2Con'
+                'leg2Con',
+                'fuel_level'
             ]
         elif self.gym_id == Gym.MCC_V0:
             names = [
@@ -436,7 +489,8 @@ class Gym(ContinuousMdpEnvironment):
             video_directory: Optional[str] = None,
             force: bool = False,
             steps_per_second: Optional[int] = None,
-            plot_environment: bool = False
+            plot_environment: bool = False,
+            progressive_reward: bool = False
     ):
         """
         Initialize the environment.
@@ -453,6 +507,7 @@ class Gym(ContinuousMdpEnvironment):
         content in the directory.
         :param steps_per_second: Number of steps per second when displaying videos.
         :param plot_environment: Whether or not to plot the environment.
+        :param progressive_reward: Use progressive reward.
         """
 
         super().__init__(
@@ -462,6 +517,7 @@ class Gym(ContinuousMdpEnvironment):
         )
 
         self.gym_id = gym_id
+        self.progressive_reward = progressive_reward
         self.continuous_action_discretization_resolution = continuous_action_discretization_resolution
         self.render_every_nth_episode = render_every_nth_episode
         if self.render_every_nth_episode is not None and self.render_every_nth_episode <= 0:
@@ -493,8 +549,8 @@ class Gym(ContinuousMdpEnvironment):
                 for i in range(self.gym_native.action_space.n)
             ]
 
-        # action space is continuous and we lack a discretization resolution:  initializee a single, multi-dimensional
-        # action including the min and max values of the dimensions.
+        # action space is continuous and we lack a discretization resolution:  initialize a single, multi-dimensional
+        # action including the min and max values of the dimensions. a policy gradient approach will be required.
         elif isinstance(self.gym_native.action_space, Box) and self.continuous_action_discretization_resolution is None:
             self.actions = [
                 ContinuousMultiDimensionalAction(
@@ -505,7 +561,7 @@ class Gym(ContinuousMdpEnvironment):
             ]
 
         # action space is continuous and we have a discretization resolution:  discretize it. this is generally not a
-        # great approach, as it results in high-dimensional action spaces.
+        # great approach, as it results in high-dimensional action spaces. but here goes.
         elif isinstance(self.gym_native.action_space, Box) and self.continuous_action_discretization_resolution is not None:
 
             box = self.gym_native.action_space
@@ -530,6 +586,13 @@ class Gym(ContinuousMdpEnvironment):
         else:  # pragma no cover
             raise ValueError(f'Unknown Gym action space type:  {type(self.gym_native.action_space)}')
 
+        # set progressive goal for certain environments
+        if self.gym_id == Gym.MCC_V0:
+            if self.progressive_reward:
+                self.mcc_curr_goal_x_pos = Gym.MCC_V0_TROUGH_X_POS + 0.1
+            else:
+                self.mcc_curr_goal_x_pos = Gym.MCC_V0_GOAL_X_POS
+
     def __getstate__(
             self
     ) -> Dict:
@@ -593,7 +656,7 @@ class CartpoleFeatureExtractor(StateActionInteractionFeatureExtractor):
             cls,
             args: List[str],
             environment: Gym
-    ) -> Tuple[StateActionInteractionFeatureExtractor, List[str]]:
+    ) -> Tuple[FeatureExtractor, List[str]]:
         """
         Initialize a feature extractor from arguments.
 
@@ -639,7 +702,7 @@ class CartpoleFeatureExtractor(StateActionInteractionFeatureExtractor):
 
         X = self.feature_scaler.scale_features(X, refit_scaler)
 
-        # interacct feature vectors per state category
+        # interact feature vectors per state category
         state_categories = [
             OneHotCategory(*[
                 obs_feature < 0.0
@@ -806,22 +869,38 @@ class ContinuousLunarLanderFeatureExtractor(ContinuousFeatureExtractor):
         # extract raw feature values
         raw_feature_values = super().extract(state, refit_scaler)
 
-        # features 0 (x pos), 2 (x velocity), 3 (y velocity), 4 (angle), and 5 (angular velocity) are signed and can be
-        # encoded categorically.
-        encoded_feature_idxs = [0, 2, 3, 4, 5]
-        feature_values_to_encode = raw_feature_values[encoded_feature_idxs]
+        # features:
+        #   0 (x pos)
+        #   1 (y pos)
+        #   2 (x velocity)
+        #   3 (y velocity)
+        #   4 (angle)
+        #   5 (angular velocity)
+        #   6 (leg 1 contact)
+        #   7 (leg 2 contact)
+        #   8 (fuel level)
+
+        # form the one-hot state category. start by thresholding some feature values.
+        state_category_feature_idxs = [0, 2, 3, 4, 5]
         state_category = OneHotCategory(*[
-            obs_feature < 0.0
-            for obs_feature in state.observation[encoded_feature_idxs]
+            value < 0.0
+            for idx, value in zip(state_category_feature_idxs, state.observation[state_category_feature_idxs])
         ])
 
+        # encode feature values
+        encoded_feature_idxs = [0, 2, 3, 4, 5]
+        feature_values_to_encode = raw_feature_values[encoded_feature_idxs]
         encoded_feature_values = self.state_category_interacter.interact(
             np.array([feature_values_to_encode]),
             [state_category]
         )[0]
 
+        # get unencoded feature values
+        both_legs_in_contact = 1.0 if all(raw_feature_values[6:8] == 1.0) else 0.0
+        unencoded_feature_values = np.append(raw_feature_values[[1, 6, 7, 8]], [both_legs_in_contact])
+
         # combine encoded and unencoded feature values
-        final_feature_values = np.append(encoded_feature_values, raw_feature_values[[1, 6, 7]])
+        final_feature_values = np.append(encoded_feature_values, unencoded_feature_values)
 
         return final_feature_values
 
@@ -841,6 +920,71 @@ class ContinuousLunarLanderFeatureExtractor(ContinuousFeatureExtractor):
         ])
 
 
+@rl_text(chapter='Feature Extractors', page=1)
+class WalkerFeatureExtractor(ContinuousFeatureExtractor):
+    """
+    Feature extractor for the walker.
+    """
+
+    def extract(
+            self,
+            state: GymState,
+            refit_scaler: bool
+    ) -> np.ndarray:
+        """
+        Extract state features.
+
+        :param state: State.
+        :param refit_scaler: Whether or not to refit the feature scaler before scaling the extracted features.
+        :return: State-feature vector.
+        """
+
+        # extract raw feature values
+        raw_feature_values = super().extract(state, refit_scaler)
+
+        # features:
+        #
+
+        # # form the one-hot state category. start by thresholding some feature values.
+        # state_category_feature_idxs = list(range(raw_feature_values.shape[0]))
+        # state_category = OneHotCategory(*[
+        #     value < 0.0
+        #     for idx, value in zip(state_category_feature_idxs, state.observation[state_category_feature_idxs])
+        # ])
+        #
+        # # encode feature values
+        # encoded_feature_idxs = state_category_feature_idxs
+        # feature_values_to_encode = raw_feature_values[encoded_feature_idxs]
+        # encoded_feature_values = self.state_category_interacter.interact(
+        #     np.array([feature_values_to_encode]),
+        #     [state_category]
+        # )[0]
+
+        # # get unencoded feature values
+        # both_legs_in_contact = 1.0 if all(raw_feature_values[6:8] == 1.0) else 0.0
+        # unencoded_feature_values = np.append(raw_feature_values[[1, 6, 7, 8]], [both_legs_in_contact])
+        #
+        # # combine encoded and unencoded feature values
+        # final_feature_values = np.append(encoded_feature_values, unencoded_feature_values)
+
+        return raw_feature_values
+
+    def __init__(
+            self
+    ):
+        """
+        Initialize the feature extractor.
+        """
+
+        super().__init__()
+
+        # interact features with relevant state categories
+        self.state_category_interacter = OneHotCategoricalFeatureInteracter([
+            OneHotCategory(*category_args)
+            for category_args in product(*([[True, False]] * 5))
+        ])
+
+
 @rl_text(chapter='Feature Extractors', page=1)
 class ContinuousMountainCarFeatureExtractor(ContinuousFeatureExtractor):
     """
@@ -865,7 +1009,7 @@ class ContinuousMountainCarFeatureExtractor(ContinuousFeatureExtractor):
 
         # encode features
         state_category = OneHotCategory(*[
-            obs_feature < Gym.MMC_V0_TROUGH_X_POS if i == 0 else  # shift the x midpoint to the trough
+            obs_feature < Gym.MCC_V0_TROUGH_X_POS if i == 0 else  # shift the x midpoint to the trough
             obs_feature <= 0.0 if i == 2 else  # fuel bottoms out at zero
             obs_feature < 0.0
             for i, obs_feature in enumerate(state.observation)
