@@ -7,8 +7,9 @@ import numpy as np
 from numpy.random import RandomState
 
 from rlai.actions import Action
+from rlai.agents import Agent
 from rlai.agents.mdp import MdpAgent, StochasticMdpAgent
-from rlai.environments.mdp import MdpEnvironment
+from rlai.environments import Environment
 from rlai.environments.network import TcpMdpEnvironment
 from rlai.meta import rl_text
 from rlai.policies import Policy
@@ -30,9 +31,9 @@ class RobocodeReward(Reward):
             r: float,
             gun_reward: float,
             movement_reward: float,
-            bullet_id_fired_event,
-            bullet_hit_events,
-            bullet_missed_events
+            bullet_id_fired_event: Dict[str, Dict],
+            bullet_hit_events: List[Dict],
+            bullet_missed_events: List[Dict]
     ):
         """
         Initialize the reward.
@@ -89,14 +90,14 @@ class RobocodeAgent(StochasticMdpAgent):
             args: List[str],
             random_state: RandomState,
             pi: Optional[Policy]
-    ) -> Tuple[StochasticMdpAgent, List[str]]:
+    ) -> Tuple[List[Agent], List[str]]:
         """
         Initialize a Robocode agent from arguments.
 
         :param args: Arguments.
         :param random_state: Random state.
         :param pi: Policy.
-        :return: 2-tuple of a Robocode agent and a list of unparsed arguments.
+        :return: 2-tuple of a list of agents and a list of unparsed arguments.
         """
 
         parsed_args, unparsed_args = parse_arguments(cls, args)
@@ -108,7 +109,7 @@ class RobocodeAgent(StochasticMdpAgent):
             **vars(parsed_args)
         )
 
-        return agent, unparsed_args
+        return [agent], unparsed_args
 
     def shape_reward(
             self,
@@ -241,7 +242,7 @@ class RobocodeEnvironment(TcpMdpEnvironment):
             cls,
             args: List[str],
             random_state: RandomState
-    ) -> Tuple[MdpEnvironment, List[str]]:
+    ) -> Tuple[Environment, List[str]]:
         """
         Initialize an environment from arguments.
 
@@ -515,8 +516,8 @@ class RobocodeState(MdpState):
             bullet_power_missed_others_cumulative: float,
             bullet_power_missed_others_since_previous_hit: float,
             events: Dict[str, List[Dict]],
-            previous_state,
-            prior_state_different_location,
+            previous_state: 'RobocodeState',
+            prior_state_different_location: 'RobocodeState',
             most_recent_scanned_robot: Optional[Dict],
             most_recent_scanned_robot_age_turns: Optional[int],
             AA: List[Action],
@@ -1450,7 +1451,7 @@ class RobocodeFeatureExtractor(FeatureExtractor):
             degrees: float
     ) -> float:
         """
-        Normalizee degress to be [0, 360].
+        Normalize degrees to be [0, 360].
 
         :param degrees: Degrees.
         :return: Normalized degrees.
