@@ -1,4 +1,5 @@
 import enum
+import logging
 from functools import partial
 from typing import Dict, Set, Tuple, Optional
 
@@ -7,9 +8,9 @@ from rlai.agents.mdp import MdpAgent
 from rlai.environments.mdp import MdpEnvironment, MdpPlanningEnvironment, PrioritizedSweepingMdpPlanningEnvironment
 from rlai.gpi import PolicyImprovementEvent
 from rlai.meta import rl_text
+from rlai.q_S_A import StateActionValueEstimator
 from rlai.states.mdp import MdpState
 from rlai.utils import IncrementalSampleAverager, sample_list_item
-from rlai.value_estimation import StateActionValueEstimator
 
 
 @rl_text(chapter=6, page=130)
@@ -67,7 +68,7 @@ def evaluate_q_pi(
     if n_steps is not None and n_steps < 1:
         raise ValueError('The value of n_steps must be in range [1, inf], or None.')
 
-    print(f'Running temporal-difference evaluation of q_pi for {num_episodes} episode(s).')
+    logging.info(f'Running temporal-difference evaluation of q_pi for {num_episodes} episode(s).')
 
     evaluated_states = set()
 
@@ -109,6 +110,8 @@ def evaluate_q_pi(
                 agent=agent
             )
 
+            logging.debug(f'Obtained reward:  {next_reward}\n')
+
             # in the case of a planning-based advancement, the planning environment returns a 3-tuple of the current
             # state, current action, and next state. this is because the planning environment may revise any one of
             # these variables to conduct the planning process (e.g., by prioritized sweeping).
@@ -124,28 +127,42 @@ def evaluate_q_pi(
             if planning_environment is not None:
                 planning_environment.model.update(curr_state, curr_a, next_state, next_reward)
 
-            # initialize the n-step, truncated return accumulator at the current time for the current state and action
+            # initialize the n-step, truncated return accumulator at the current time for the current state and action.
             t_state_a_g[curr_t] = (curr_state, curr_a, 0.0)
 
-            # get prior time steps for which truncated return g should be updated. if n_steps is None, then get all
-            # prior time steps (equivalent to infinite n_steps, or Monte Carlo).
+            # ask the agent to shape the reward, returning the time steps whose returns should be updated and the shaped
+            # reward associated with each. if n_steps is None, then shape the reward all the way back to the start
+            # (equivalent to infinite n_steps, or monte carlo returns). if n_steps is not None, then shape the reward
+            # for n-step updates.
             if n_steps is None:
-                prior_t_values = list(t_state_a_g.keys())
+                first_t = 0
             else:
-                # in 1-step td, the earliest time step is the current time step; in 2-step, the earliest time step is
-                # the prior time step, etc. always update through the current time step.
-                earliest_t = max(0, curr_t - n_steps + 1)
-                prior_t_values = range(earliest_t, curr_t + 1)
+                # in 1-step td, the earliest time step is the final time step; in 2-step, the earliest time step is the
+                # prior time step, etc.
+                first_t = max(0, curr_t - n_steps + 1)
 
-            # pass reward to prior state-action values, discounting based on time step differences (the reward should
-            # not be discounted for the current time step).
-            for t in prior_t_values:
-                state, a, g = t_state_a_g[t]
-                discount = agent.gamma ** (curr_t - t)
-                t_state_a_g[t] = (state, a, g + discount * next_reward.r)
+            t_shaped_reward = agent.shape_reward(
+                reward=next_reward,
+                first_t=first_t,
+                final_t=curr_t
+            )
+
+            # update return accumulators with shaped rewards
+            t_state_a_g.update({
+
+                return_t: (
+                    t_state_a_g[return_t][0],
+                    t_state_a_g[return_t][1],
+                    t_state_a_g[return_t][2] + shaped_reward
+                )
+                for return_t, shaped_reward in t_shaped_reward.items()
+
+                # reward shapers might return invalid time steps. ignore these.
+                if return_t in t_state_a_g
+            })
 
             # get the next state's bootstrapped value and next action, based on the bootstrapping mode. note that the
-            # bootstrapped next state-action value is only used if we're performing n-step updates.
+            # bootstrapped next state-action value is only used if we're performing n-step updates below.
             next_state_q_s_a, next_a = get_bootstrapped_state_action_value(
                 state=next_state,
                 t=next_t,
@@ -180,8 +197,8 @@ def evaluate_q_pi(
         # flush out the remaining n-step updates. if we terminated because we reached a terminal state, then all next-
         # state values for the updates are zero. if instead we terminated because we reached the maximum number of time
         # steps, then use the bootstrapped next-state value instead. this is an important distinction because these
-        # value can be dramatically different, and when discounting is not using (or is very small), the resulting
-        # value estimates can be correspondingly different.
+        # value can be dramatically different, and when discounting is not used (or is very small), the resulting value
+        # estimates can be correspondingly different.
         if curr_state.terminal:
             next_state_q_s_a = 0.0
 
@@ -205,7 +222,7 @@ def evaluate_q_pi(
 
         episodes_finished = episode_i + 1
         if episodes_finished % episodes_per_print == 0:
-            print(f'Finished {episodes_finished} of {num_episodes} episode(s).')
+            logging.info(f'Finished {episodes_finished} of {num_episodes} episode(s).')
 
     return evaluated_states, episode_reward_averager.get_value()
 
@@ -246,7 +263,7 @@ def get_bootstrapped_state_action_value(
         else:
 
             # SARSA:  agent determines the t-d target action as well as the episode's next action, which are the same
-            # (we're on-policy)
+            # (we're on-policy).
             if mode == Mode.SARSA:
                 td_target_a = next_a = agent.act(t)
 
@@ -266,7 +283,7 @@ def get_bootstrapped_state_action_value(
             else:
                 bootstrapped_s_a_value = 0.0
 
-        # if we're off-policy, then we won't yet have a next action. ask the agent for it now.
+        # if we're off-policy, then we won't yet have a next action. ask the agent for an action now.
         if next_a is None:
             next_a = agent.act(t)
 
@@ -293,7 +310,7 @@ def update_q_S_A(
     :param q_S_A: State-action value estimator.
     :param n_steps: Number of time steps to accumulate actual rewards for before updating a state-action value.
     :param curr_t: Current time step.
-    :param t_state_a_g: Structore of time, state, action, g accumulators. If an n-step update is feasible at the current
+    :param t_state_a_g: Structure of time, state, action, g accumulators. If an n-step update is feasible at the current
     time step (i.e., if we have accumulated sufficient rewards), then the entry corresponding to the n-step update will
     be deleted from this structure.
     :param agent: Agent.
