@@ -155,6 +155,23 @@ class ContinuousActionPolicy(ParameterizedPolicy, ABC):
 
         return True
 
+    def __getstate__(
+            self
+    ) -> Dict:
+        """
+        Get state dictionary for pickling.
+
+        :return: State dictionary.
+        """
+
+        state = dict(self.__dict__)
+
+        # the environment should not be pickled, as it won't be used when running or resuming the policy later, and it
+        # will contain references (e.g., to scatter plots) that will become invalid.
+        state['environment'] = None
+
+        return state
+
 
 @rl_text(chapter=13, page=335)
 class ContinuousActionNormalDistributionPolicy(ContinuousActionPolicy):
@@ -287,7 +304,7 @@ class ContinuousActionNormalDistributionPolicy(ContinuousActionPolicy):
             # off-diagonal elements can be positive or negative
             else np.dot(theta_cov_row, state_features)
 
-            # iteraate over each row of coefficients
+            # iterate over each row of coefficients
             for i, theta_cov_row in enumerate(theta_cov)
 
         ]).reshape(self.environment.get_action_space_dimensionality(), self.environment.get_action_space_dimensionality())
@@ -320,7 +337,7 @@ class ContinuousActionNormalDistributionPolicy(ContinuousActionPolicy):
             # off-diagonal elements can be positive or negative
             else jnp.dot(theta_cov_row, state_features)
 
-            # iteraate over each row of coefficients
+            # iterate over each row of coefficients
             for i, theta_cov_row in enumerate(theta_cov)
 
         ]).reshape(action_space_dimensionality, action_space_dimensionality)
@@ -410,7 +427,7 @@ class ContinuousActionNormalDistributionPolicy(ContinuousActionPolicy):
         :return: State dictionary.
         """
 
-        state = dict(self.__dict__)
+        state = super().__getstate__()
 
         state['get_action_density_gradients'] = None
         state['get_action_density_gradients_vmap'] = None
@@ -434,7 +451,7 @@ class ContinuousActionNormalDistributionPolicy(ContinuousActionPolicy):
 
     def __eq__(
             self,
-            other
+            other: object
     ) -> bool:
         """
         Check whether the current policy equals another.
@@ -443,14 +460,15 @@ class ContinuousActionNormalDistributionPolicy(ContinuousActionPolicy):
         :return: True if policies are equal and False otherwise.
         """
 
-        other: ContinuousActionNormalDistributionPolicy
+        if not isinstance(other, ContinuousActionNormalDistributionPolicy):
+            raise ValueError(f'Expected {ContinuousActionNormalDistributionPolicy}')
 
         # using the default values for allclose is too strict to achieve cross-platform testing success. back off a little with atol.
         return np.allclose(self.theta_mean, other.theta_mean, atol=0.0001) and np.allclose(self.theta_cov, other.theta_cov, atol=0.0001)
 
     def __ne__(
             self,
-            other
+            other: object
     ) -> bool:
         """
         Check whether the current policy does not equal another.
@@ -558,23 +576,30 @@ class ContinuousActionBetaDistributionPolicy(ContinuousActionPolicy):
                 # TODO:  How to estimate this from the PDF?
                 p_a_s = 1.0
 
-                # check for nans in the gradients and skip the update if any are found
-                if np.isnan(action_density_gradient_wrt_theta_a).any() or np.isnan(action_density_gradient_wrt_theta_b).any():  # pragma no cover
-                    warnings.warn('Gradients contain np.nan value(s). Skipping update.')
-                else:
+                # squash gradients into [-1.0, 1.0] to handle scaling issues when actions are chosen at the tails of
+                # the beta distribution where the gradients are very large. this also addresses cases of
+                # positive/negative infinite gradients.
 
-                    # squash gradients into [-1.0, 1.0] to handle scaling issues when actions are chosen at the tails of
-                    # the beta distribution where the gradients are very large. this also addresses cases of
-                    # positive/negative infinite gradients.
-                    self.action_theta_a[action_i, :] += alpha * target * (np.tanh(action_density_gradient_wrt_theta_a) / p_a_s)
-                    self.action_theta_b[action_i, :] += alpha * target * (np.tanh(action_density_gradient_wrt_theta_b) / p_a_s)
+                action_density_gradient_wrt_theta_a = np.nan_to_num(action_density_gradient_wrt_theta_a)
+                self.action_theta_a[action_i, :] += alpha * target * (np.tanh(action_density_gradient_wrt_theta_a) / p_a_s)
+
+                action_density_gradient_wrt_theta_b = np.nan_to_num(action_density_gradient_wrt_theta_b)
+                self.action_theta_b[action_i, :] += alpha * target * (np.tanh(action_density_gradient_wrt_theta_b) / p_a_s)
 
         self.reset_action_scatter_plot_y_range()
 
+        if logging.getLogger().level <= logging.DEBUG:
+            action_min_a = np.amin(self.action_theta_a, 1)
+            action_max_a = np.amax(self.action_theta_a, 1)
+            action_min_b = np.amin(self.action_theta_b, 1)
+            action_max_b = np.amax(self.action_theta_b, 1)
+            for i, (min_a, max_a, min_b, max_b) in enumerate(zip(action_min_a, action_max_a, action_min_b, action_max_b)):
+                logging.debug(f'Action {i} [min,max]:\n\ta:  [{min_a},{max_a}]\n\tb:  [{min_b},{max_b}]\n')
+
         # only output the hyperparameter table if we have a state-dimension name for each feature. some feature
-        # extractors expand the feature space beyond the dimensions, and we don't have a good way to generate names for
-        # these extra dimensions. such extractors also tend to generate large feature spaces for which tabular output
-        # isn't readable.
+        # extractors expand the feature space beyond the dimension names (e.g., via interaction terms), and we don't
+        # have a good way to generate names for these extra features. such extractors also tend to generate large
+        # feature spaces for which tabular output isn't easily readable.
         if logging.getLogger().level <= logging.DEBUG and self.action_theta_a.shape[1] == len(self.environment.get_state_dimension_names()) + 1:
             row_names = [
                 f'{action_name}_{p}'
@@ -780,7 +805,7 @@ class ContinuousActionBetaDistributionPolicy(ContinuousActionPolicy):
         :return: State dictionary.
         """
 
-        state = dict(self.__dict__)
+        state = super().__getstate__()
 
         state['get_action_density_gradients'] = None
         state['get_action_density_gradients_vmap'] = None
@@ -804,7 +829,7 @@ class ContinuousActionBetaDistributionPolicy(ContinuousActionPolicy):
 
     def __eq__(
             self,
-            other
+            other: object
     ) -> bool:
         """
         Check whether the current policy equals another.
@@ -813,14 +838,15 @@ class ContinuousActionBetaDistributionPolicy(ContinuousActionPolicy):
         :return: True if policies are equal and False otherwise.
         """
 
-        other: ContinuousActionBetaDistributionPolicy
+        if not isinstance(other, ContinuousActionBetaDistributionPolicy):
+            raise ValueError(f'Expected {ContinuousActionBetaDistributionPolicy}')
 
         # using the default values for allclose is too strict to achieve cross-platform testing success. back off a little with atol.
         return np.allclose(self.action_theta_a, other.action_theta_a, atol=0.0001) and np.allclose(self.action_theta_b, other.action_theta_b, atol=0.0001)
 
     def __ne__(
             self,
-            other
+            other: object
     ) -> bool:
         """
         Check whether the current policy does not equal another.
