@@ -1,20 +1,21 @@
 import sys
 from argparse import ArgumentParser
-from typing import List
+from typing import List, Union, Tuple, Optional
 
 from rlai.meta import rl_text
 from rlai.runners import trainer, agent_in_environment
-from rlai.utils import load_class, parse_arguments
+from rlai.utils import parse_arguments, get_argument_parser
 
 
 @rl_text(chapter='Training and Running Agents', page=1)
 def run(
         args: List[str] = None
-):
+) -> Union[None, Tuple[Optional[str], str]]:
     """
     Run RLAI.
 
     :param args: Arguments.
+    :return: Return value of specified function.
     """
 
     # create the top-level rlai parser and add subparsers for commands
@@ -34,7 +35,7 @@ def run(
     help_parser.set_defaults(func=show_help)
 
     parsed_args, unparsed_args = parser.parse_known_args(args)
-    parsed_args.func(unparsed_args)
+    return parsed_args.func(unparsed_args)
 
 
 def show_help(
@@ -46,16 +47,15 @@ def show_help(
     :param args: Arguments.
     """
 
-    if len(args) != 1:
+    if len(args) == 1:
+        try:
+            parser = get_argument_parser(args[0])
+            parse_arguments(parser, ['--help'])
+        except Exception as ex:
+            print(f'{ex}')
+    else:
         print('Usage:  rlai help CLASS')
 
-    try:
-        loaded_class = load_class(args[0])
-        parser = loaded_class.get_argument_parser()
-        parse_arguments(parser, ['--help'])
-    except Exception as ex:
-        print(f'{ex}')
 
-
-if __name__ == '__main__':
+if __name__ == '__main__':  # pragma no cover
     run(sys.argv[1:])
