@@ -1,5 +1,6 @@
 import importlib
 import logging
+import os
 import threading
 from argparse import Namespace, ArgumentParser
 from enum import Enum, auto
@@ -25,6 +26,10 @@ class IncrementalSampleAverager:
     def reset(
             self
     ):
+        """
+        Reset the average and number of samples.
+        """
+
         self.average = self.initial_value
         self.n = 0
 
@@ -113,22 +118,42 @@ class IncrementalSampleAverager:
     def __str__(
             self
     ) -> str:
+        """
+        Get string.
+
+        :return: String.
+        """
 
         return str(self.average)
 
     def __eq__(
             self,
-            other
+            other: object
     ) -> bool:
+        """
+        Check equality with another averager.
+
+        :param other: Averager.
+        :return: True if average values match and False otherwise.
+        """
+
+        if not isinstance(other, IncrementalSampleAverager):
+            raise ValueError(f'Expected {IncrementalSampleAverager}')
 
         return self.get_value() == other.get_value()
 
     def __ne__(
             self,
-            other
+            other: object
     ) -> bool:
+        """
+        Check inequality with another averager.
 
-        return self.get_value() != other.get_value()
+        :param other: Averager.
+        :return: True if average values do not match and False otherwise.
+        """
+
+        return not (self == other)
 
 
 def sample_list_item(
@@ -187,7 +212,7 @@ def import_function(
 
 def load_class(
         fully_qualified_class_name: str
-):
+) -> Any:
     """
     Load class from its fully-qualified name (e.g., xxx.yyy.Class).
 
@@ -203,7 +228,7 @@ def load_class(
 
 
 def get_argument_parser(
-        fully_qualified_class_name
+        fully_qualified_class_name: str
 ) -> ArgumentParser:
     """
     Get argument parser for a class.
@@ -254,7 +279,7 @@ def get_base_argument_parser(
 
 
 def parse_arguments(
-        parser_or_cls,
+        parser_or_cls: Any,
         args: List[str]
 ) -> Tuple[Namespace, List[str]]:
     """
@@ -389,7 +414,9 @@ def log_with_border(
         logging.log(level, '')
 
 
-def get_nearest_positive_definite_matrix(A):
+def get_nearest_positive_definite_matrix(
+        A: np.ndarray
+) -> np.ndarray:
     """
     Find the nearest positive-definite matrix to input. A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code
     [1], which credits [2].
@@ -426,7 +453,7 @@ def get_nearest_positive_definite_matrix(A):
     # will be much larger than [1]'s `eps(mineig)`, since `mineig` is usually on
     # the order of 1e-16, and `eps(1e-16)` is on the order of 1e-34, whereas
     # `spacing` will, for Gaussian random matrixes of small dimension, be on
-    # othe order of 1e-16. In practice, both ways converge, as the unit test
+    # the order of 1e-16. In practice, both ways converge, as the unit test
     # below suggests.
     eye = np.eye(A.shape[0])
     k = 1
@@ -439,7 +466,9 @@ def get_nearest_positive_definite_matrix(A):
     return A3
 
 
-def is_positive_definite(m):
+def is_positive_definite(
+        m: np.ndarray
+) -> bool:
     """
     Returns true when input is positive-definite, via Cholesky.
 
@@ -658,3 +687,21 @@ class ScatterPlot:
         plot_x_axis.setTicks([list(enumerate(self.x_tick_labels))])
         self.plot_widget = self.plot_layout.addPlot(axisItems={'bottom': plot_x_axis})
         self.set_position(self.position)
+
+
+def insert_index_into_path(
+        path: str,
+        index: int
+) -> str:
+    """
+    Insert an index into a path.
+
+    :param path: Path.
+    :param index: Index.
+    :return: Path with index.
+    """
+
+    path_parts = [p for p in os.path.splitext(path) if p != ""]
+    path_parts.insert(1, f'-{index}')
+
+    return ''.join(path_parts)
